#include <iostream>
#include <fstream>

#include <boost/lexical_cast.hpp>
#include "CodeGenerator.h"

using namespace boost;
using namespace std;
using namespace Compiler;

//*******************************************************************
//*******************************************************************
//*******************************************************************

// Modify the instruction to its local version if the scope in which it is
// generated is local.
//
void Generator::checkInstrScope(SymbolicInstr* instr) {
	if ((currentFunction || instr->scope == SymbolicInstr::LOCAL) &&
			instr->cspec != SymbolicInstr::STANDARD) {
		for (int i = 0;  i < CODE_SCOPE_TABLE_SIZE;  ++i) {
			// check if the instruction has a local version ...
			if (instr->opcode == codeScopeTable[i][0]) {
				// ... if it does it's a LOAD/STORE/INC/DEC instruction which
				// could have a symbol as an operand (the operand can also be a
				// constant); if a symbol is the operand and the symbol is
				// non-local then use the non-local version of the instruction.
				//
				if (instr->opnd1 && !instr->opnd1->isLocal) return;

				// use the local version
				instr->opcode = codeScopeTable[i][1];
				break;
			}
		}
	}
}

//*******************************************************************

void Generator::genAsm(SymbolicInstr instr, int pos) {
	checkInstrScope(&instr);
	
	if (pos > -1) {
		code[pos] = instr;
	}
	else {
		code.push_back(instr);
	}

	string mnemonic = getMnemonic(instr.opcode);
}

//*******************************************************************

void Generator::removeLastInstr() {
	code.erase(code.end() - 1);
}

//*******************************************************************

void Generator::genConst(const Constant& cns) {
	if (cns.type.getName() == INT || cns.type.getName() == BOOL || cns.type.getName() == CHAR) {
		genAsm(SymbolicInstr(ICLOAD, cns));
	}
	else {
		genAsm(SymbolicInstr(LOAD, cns, cns.type.getStorageClass()));
	}
}

//*******************************************************************

TypeAttr Generator::genAddressOf(Symbol* sym) {
	if (sym->isLocal) {
		genAsm(SymbolicInstr(LOADSP));
		genAsm(SymbolicInstr(ICLOAD, sym));
		genAsm(SymbolicInstr(ADD, INT_CODE));
	}
	else {
		genAsm(SymbolicInstr(ICLOAD, sym));
	}
	
	TypeAttr t = sym->type;
	t.setAsPointer();
	return t;
}

//*******************************************************************

void Generator::backpatchAt(int instrLoc, Constant c) {
	SymbolicInstr* ip = &(code[instrLoc]);
	Constant* p = &(ip->cns);
	*p = c;
}

//*******************************************************************

void Generator::backpatch(Symbol* sym, BackpatchTime when, int n) {
	for (BackpatchVector::iterator itr = bp.begin();  itr != bp.end();  ++itr) {
		IBackpatch* bpi = *itr;
		if (bpi->sym == sym && bpi->when == when) {
			backpatchAt(bpi->pos, Constant(bpi->value(n)));
			delete bpi;
			bp.erase(itr);
			--itr;
		}
	}
}

//*******************************************************************

void Generator::patchFunctionCalls(SymbolTable* st) {
	for (SymbolTable::SymbolPtrMap::const_iterator symItr = st->getFnTable().begin();
			symItr != st->getFnTable().end();
			++symItr) {

		Symbol* sym = symItr->second;
		
		if (sym->isFunction) {
			for (FunctionCallVector::iterator itr = fnCall.begin();  itr != fnCall.end();  ++itr) {
				int v = 0;
				
				if ((*itr)->name == sym->name) {
					if ((*itr)->resultSpace > 0) {
						backpatchAt((*itr)->resultSpace, Constant(sym->type.isArray() ? 1 : sym->type.getSize(WORD, TypeAttr::ARRAY_AS_POINTER)));
					}

					if ((*itr)->copyResult > 0) {
						// Here we need to get to the returned value which is
						// currently on the expression stack of the called function.
						
						// add 2 for one word for return address + one word for SP
						v = sym->localSpace + sym->type.getSize(WORD, TypeAttr::ARRAY_AS_POINTER) + 2;

						backpatchAt((*itr)->copyResult, Constant(v * WORD_SIZE));
					}
					
					v = sym->localSpace - sym->paramsSpace;
					if (v > 0) {
						backpatchAt((*itr)->nonParamSpace, Constant(v));
					}
					
					backpatchAt((*itr)->targetAddr, Constant(sym->index));
					
					// The DECX instruction needs to set the expression stack
					// pointer to the address where the result value will be copied.
					//
					v =	sym->type.getSize(WORD, TypeAttr::ARRAY_AS_POINTER) +	// the value on the current (local) expression stack
							2 +																	// one word for return address + one word for SP
							sym->localSpace +													// local data
							sym->type.getSize(WORD, TypeAttr::ARRAY_AS_POINTER);	// the space reserved for the result value
								
					backpatchAt((*itr)->restoreX, Constant(v));
					
					// The SP is restored by setting it to its previous address
					// which is stored in its frame right bellow the local
					// expression stack pointer. Since the expression stack pointer
					// is already restored the SP address cannot be obtained by
					// calculating XSP - 1.
					//
					v =	sym->type.getSize(WORD, TypeAttr::ARRAY_AS_POINTER) +	// the result value
							sym->localSpace +													// local data
							1;																		// return address
					backpatchAt((*itr)->restoreSP, Constant(v * WORD_SIZE));
					
					delete *itr;
					fnCall.erase(itr);
					--itr;
				}
			}
		}
	}
}

//*******************************************************************

void Generator::setArrayLayoutValues(const TypeAttr& arrayType, vector<int>* v) {
	int n = 1;
	
	v->push_back(n);
	for (size_t i = arrayType.getArrayDim().size() - 1;  i > 0;  --i) {
		n *= arrayType.getArrayDim()[i];
		v->push_back(n);
	}
}

//*******************************************************************

void Generator::generateAssembly(string programName, Scope::SymbolTableMap& allSymbolTables) {
	string debugFileName = inputArgs[INPUT_PATH_SWITCH] + programName + string(DEBUG_FILE_EXT);
	ofstream debugFile(debugFileName.c_str());

	debugFile << "*** CODE ***\n\n";

	int i = 0;
	for (InstrVector::iterator itr = code.begin();  itr != code.end();  ++itr, ++i) {
		SymbolicInstr instr = *itr;
		debugFile << i << ": " << instr << endl;
	}
	
	debugFile << "\n\n";

	string asmFileName = inputArgs[INPUT_PATH_SWITCH] + programName + string(OUTPUT_FILE_EXT);
	ofstream asmFile(asmFileName.c_str());

	generateDataHeader(allSymbolTables, asmFile, debugFile);

	asmFile << endl << CODE_SEGMENT << endl;
	for (InstrVector::iterator itr = code.begin();  itr != code.end();  ++itr) {
		SymbolicInstr instr = *itr;

		string s = mnemonic[instr.opcode];

		if (instr.cspec != SymbolicInstr::NONE) {
			if (instr.argCount > 0) {
				if (instr.cns.type.getName() == INT || instr.cns.type.getName() == BOOL || 
					(instr.cns.type.getName() == CHAR && !instr.cns.type.isArray())) {
					s += " " + instr.cns.token;
					
					if (instr.argCount > 1) {
						s += ", " + instr.auxCns.token;
					}
				}
				else {
					Constant c = SymbolTable::getConstant(instr.cns.token);
					s += " " + lexical_cast<string>(c.index);
				}
			}
		}
		else {
			if (instr.argCount > 0) {
				Symbol* sym = instr.opnd1;
				s += " " + lexical_cast<string>(Compiler::findSymbol(sym->name, sym->symbolTableID, allSymbolTables)->index);
			}
		}

		asmFile << s << endl;
	}

	asmFile << endl << ASM_END << endl;
}

//*******************************************************************

TypeAttr Generator::generateArithmeticInstr(Operation optr, const TypeAttr& leftOpnd, const TypeAttr& rightOpnd, int loc) {
	// generate ITOF if one of the operands is float
	if (leftOpnd != rightOpnd && (leftOpnd == TypeAttr(FLOAT) || rightOpnd == TypeAttr(FLOAT))) {
		TypeAttr type(FLOAT);

		if (leftOpnd == TypeAttr(INT)) {
			// this needs to replace the NOP instruction
			genAsm(SymbolicInstr(Compiler::ITOF), loc);
		}
		else if (rightOpnd == TypeAttr(INT)) {
			genAsm(SymbolicInstr(Compiler::ITOF));
		}

		genAsm(SymbolicInstr(optr, type.getTypeNameCode()));

		return type;
	}
	else {	// either both char/int or both float
		genAsm(SymbolicInstr(optr, leftOpnd.getTypeNameCode())); 
				
		return leftOpnd;
	}
}

//*******************************************************************

bool Compiler::validateFunctionCall(Symbol* sym, vector<TypeAttr>* args) {
	if (sym->isForwardedFunction) {
		if (args->size() != sym->fwdParamTypes.size()) {
			return false;
		}
	}
	else if (args->size() != sym->params.size()) {
		return false;
	}

	size_t paramIndex = 0;
	for (vector<TypeAttr>::iterator itr = args->begin();  itr != args->end();  ++itr, ++paramIndex) {
		TypeAttr argType = *itr;

		if (sym->isForwardedFunction) {
			TypeAttr paramType = sym->fwdParamTypes[paramIndex];
			
			if (! argType.isAssignableTo(paramType)) {
				return false;
			}
		}
		else {
			TypeAttr paramType = sym->params[paramIndex]->type;
			
			if (! argType.isAssignableTo(paramType)) {
				return false;
			}
		}
	}
	
	return true;
}

//*******************************************************************

ostream& Compiler::operator<<(ostream& os, const Instruction& instr) {
	string s = mnemonic[instr.opcode];
	if (instr.opnd1 > 0) s += " " + lexical_cast<string>(instr.opnd1);

	return os << s;
}

//*******************************************************************

ostream& Compiler::operator<<(ostream& os, const SymbolicInstr& instr) {
	string s = mnemonic[instr.opcode];

	if (instr.cspec != SymbolicInstr::NONE) {
		s += " <" + instr.cns.token + ">";
		if (instr.argCount > 1) s += ", <" + instr.auxCns.token + ">";
	}
	else {
		if (instr.argCount > 0) s += " \"" + instr.opnd1->name + "\"";
	}

	return os << s;
}
