/*
	Purpose	:	This file contains some helper functions for code generation in AST.

	Author	:	Alex Stojan
*/

#ifndef _CODEGENERATOR_
#define _CODEGENERATOR_

#include <vector>
#include <string>

#include "IC.h"
#include "Globals.h"
#include "Symbols.h"

namespace Compiler {
	struct Instruction {
		Instruction(int rator, int rand1 = 0, int rand2 = 0):
			opcode(rator),
			opnd1(rand1),
			opnd2(rand2) {}

		int opcode;
		int opnd1;
		int opnd2;
	};

	struct SymbolicInstr {
		enum ConstSpec { STANDARD, OPERAND, NONE };
		enum Scope { LOCAL, DEPENDENT };
		
		explicit SymbolicInstr(int rator, int typeCode = UNDEF_TYPE, Scope s = DEPENDENT):
			cspec(NONE),
			scope(s),
			opcode(codeTypeTable[rator][typeCode == STRING_CODE ? CHAR_CODE : typeCode]),
			opnd1(0),
			argCount(0) {}

		SymbolicInstr(int rator, Symbol* rand1, int typeCode = UNDEF_TYPE, Scope s = DEPENDENT):
			cspec(NONE),
			scope(s),
			opcode(codeTypeTable[rator][typeCode == STRING_CODE ? CHAR_CODE : typeCode]),
			opnd1(rand1),
			argCount(1) {}

		SymbolicInstr(int rator, Constant c, int typeCode = UNDEF_TYPE, ConstSpec spec = STANDARD, Scope s = DEPENDENT):
			cspec(spec),
			scope(s),
			opcode(codeTypeTable[rator][typeCode == STRING_CODE ? CHAR_CODE : typeCode]),
			opnd1(0),
			cns(c),
			argCount(1) {}

		// Used for instructions like SYSCALL that just have a couple of code 
		// values as arguments that are not tied to any identifiers in the source
		// code.
		//
		SymbolicInstr(int rator, Constant c1, Constant c2, int typeCode = UNDEF_TYPE):
			cspec(OPERAND),
			opcode(codeTypeTable[rator][typeCode == STRING_CODE ? CHAR_CODE : typeCode]),
			opnd1(0),
			cns(c1),
			auxCns(c2),
			argCount(2) {}
				
		ConstSpec	cspec;
		Scope			scope;
		int			opcode;
		Symbol*		opnd1;
		Constant		cns;
		Constant		auxCns;
		int			argCount;
	};

	enum BackpatchTime { AFTER_FN_BODY };
	
	// This structure is used for backpatching all instructions for which patch
	// value cannot be determined locally. The value(int) method specifies how
	// the patch value is calculated.
	//
	struct IBackpatch {
		IBackpatch(int p, int v, Symbol* s, BackpatchTime w):
			pos(p), 
			initial(v), 
			sym(s),
			when(w) {}
		
		virtual int value(int) const = 0;
		
		int		pos;
		int		initial;
		Symbol*	sym;
		int		when;
	};

	struct FunctionCall {
		FunctionCall():
			resultSpace(0),
			nonParamSpace(0),
			targetAddr(0),
			restoreX(0),
			copyResult(0),
			restoreSP(0) {}
		
		std::string name;
		int resultSpace;
		int nonParamSpace;
		int targetAddr;
		int restoreX;
		int copyResult;
		int restoreSP;
	};

	// This structure contains some common methods used when generating code.
	// There is only one object of this type.
	//
	struct Generator {
		typedef std::vector<SymbolicInstr>	InstrVector;
		typedef std::vector<IBackpatch*>		BackpatchVector;
		typedef std::vector<FunctionCall*>	FunctionCallVector;

		Symbol*					currentFunction;
		bool						hasReturnStatement;
		InstrVector				code;
		BackpatchVector		bp;
		FunctionCallVector	fnCall;
		
		Generator():
			currentFunction(0),
			hasReturnStatement(false) {}
		
		std::string getMnemonic(int code) { return mnemonic[code]; }
		void checkInstrScope(SymbolicInstr*);
		void genAsm(SymbolicInstr, int = -1);
		void removeLastInstr();
		OpCode getTypedInstruction(const TypeAttr&);
		void setArrayLayoutValues(const TypeAttr&, std::vector<int>*);
		void generateAssembly(std::string, Scope::SymbolTableMap&);
		void genConst(const Constant&);
		TypeAttr genAddressOf(Symbol*);
		int getCodeSize() const { return static_cast<int>(code.size()); }
		void backpatchAt(int, Constant);
		void backpatch(Symbol*, BackpatchTime, int);
		void patchFunctionCalls(SymbolTable*);
		TypeAttr generateArithmeticInstr(Operation, const TypeAttr&, const TypeAttr&, int = -1);
	};
	
	bool validateFunctionCall(Symbol*, std::vector<TypeAttr>*);

	std::ostream& operator<<(std::ostream&, const Instruction&);
	std::ostream& operator<<(std::ostream&, const SymbolicInstr&);
}

#endif
