/*
	Purpose	:	Global constants and definitions.
	
	Author	:	Alex Stojan
*/

#ifndef _GLOBALS_
#define _GLOBALS_

#include <string>
#include <map>

#include "IC.h"

namespace Compiler {
	const int WORD_SIZE	= 4;

	// system codes for SYSCALL instruction
	const int SYSCODE_WRITE		= 1;
	const int SYSCODE_READ		= 2;
	const int SYSCODE_WRITELN	= 3;

	// type symbols
	char* const CHAR		= "char";
	char* const BOOL		= "bool";
	char* const INT		= "int";
	char* const FLOAT		= "float";
	char* const VOID		= "void";
	
	// numeric type codes
	enum TypeCode {
		STRING_CODE,
		CHAR_CODE,
		BOOL_CODE,
		INT_CODE,
		FLOAT_CODE,
		VOID_CODE,
		UNDEF_TYPE
	};
	
	enum Operation {
		NOP,
		LOAD,
		LOADI,
		LOADI0,
		STORE,
		STOREI,
		STOREI0,
		ICLOAD,
		ADD,
		SUB,
		MUL,
		DIV,
		MOD,
		NEG,
		AND,
		OR,
		NOT,
		LT,
		GT,
		LE,
		GE,
		EQ,
		NEQ,
		INC1,
		DEC1,
		INCI,
		DECI,
		ITOF,
		FTOI,
		INCX,
		DECX,
		GOTO,
		IFTRUE,
		IFFALSE,
		GOTOI,
		PUSHX,
		POPXSP,
		LOADSP,
		SETSP,
		SYSCALL
	};
	
	// This table is used for getting the right instruction based on a type.
	// The order of the table columns corresponds to the order in the TypeCode
	// enumeration. When SymbolicInstr object is created the default type is
	// UNDEF_TYPE, which has value 6. This means that the last column (denoted
	// with n/a) will be used to get the instruction code based on the operator
	// field (0th column). For example, if the type is UNDEF_TYPE and the
	// operator is GOTO, the SymbolicInstr constructor will set its opcode value
	// to IC_GOTO, stored in the table at codeTypeTable[GOTO][UNDEF_TYPE]. If
	// the type is float and the operator is ADD the opcode will be stored at
	// [ADD][FLOAT], which is IC_FADD. If an instruction is not type-dependent,
	// the table contains 0 at all type locations, except void and undef_type.
	// This would cause the compiler to generate NOP instructions if a wrong
	// table entry is used.
	//
	const int codeTypeTable[][40] = {
	//		oper.			char			bool			int			float			void			n/a
		{	NOP,			0,				0,				0,				0,				IC_NOP,		IC_NOP		},
		{	LOAD,			IC_BLOAD,	IC_BLOAD,	IC_WLOAD,	IC_WLOAD,	0,				0				},
		{	LOADI,		IC_BLOADI,	IC_BLOADI,	IC_WLOADI,	IC_WLOADI,	0,				0				},
		{	LOADI0,		IC_BLOADIZ,	IC_BLOADIZ,	IC_WLOADIZ,	IC_WLOADIZ,	0,				0				},
		{	STORE,		IC_BSTORE,	IC_BSTORE,	IC_WSTORE,	IC_WSTORE,	0,				0				},
		{	STOREI,		IC_BSTOREI,	IC_BSTOREI,	IC_WSTOREI,	IC_WSTOREI,	0,				0				},
		{	STOREI0,		IC_BSTOREIZ,IC_BSTOREIZ,IC_WSTOREIZ,IC_WSTOREIZ,0,				0				},
		{	ICLOAD,		0,				0,				0,				0,				IC_ICLOAD,	IC_ICLOAD	},
		{	ADD,			IC_IADD,		IC_IADD,		IC_IADD,		IC_FADD,		0,				0				},
		{	SUB,			IC_ISUB,		IC_ISUB,		IC_ISUB,		IC_FSUB,		0,				0				},
		{	MUL,			IC_IMUL,		IC_IMUL,		IC_IMUL,		IC_FMUL,		0,				0				},
		{	DIV,			IC_IDIV,		IC_IDIV,		IC_IDIV,		IC_FDIV,		0,				0				},
		{	MOD,			0,				0,				0,				0,				IC_MOD,		IC_MOD		},
		{	NEG,			IC_INEG,		IC_INEG,		IC_INEG,		IC_FNEG,		0,				0				},
		{	AND,			0,				0,				0,				0,				IC_AND,		IC_AND		},
		{	OR,			0,				0,				0,				0,				IC_OR,		IC_OR			},
		{	NOT,			0,				0,				0,				0,				IC_NOT,		IC_NOT		},
		{	LT,			IC_ICMPLT,	IC_ICMPLT,	IC_ICMPLT,	IC_FCMPLT,	0,				0				},
		{	GT,			IC_ICMPGT,	IC_ICMPGT,	IC_ICMPGT,	IC_FCMPGT,	0,				0				},
		{	LE,			IC_ICMPLE,	IC_ICMPLE,	IC_ICMPLE,	IC_FCMPLE,	0,				0				},
		{	GE,			IC_ICMPGE,	IC_ICMPGE,	IC_ICMPGE,	IC_FCMPGE,	0,				0				},
		{	EQ,			IC_WCMPEQ,	IC_WCMPEQ,	IC_WCMPEQ,	IC_WCMPEQ,	0,				0				},
		{	NEQ,			IC_WCMPNEQ,IC_WCMPNEQ,IC_WCMPNEQ,IC_WCMPNEQ,0,				0				},
		{	INC1,			0,				0,				0,				0,				IC_INC1,		IC_INC1		},
		{	DEC1,			0,				0,				0,				0,				IC_DEC1,		IC_DEC1		},
		{	INCI,			0,				0,				0,				0,				IC_INCI,		IC_INCI		},
		{	DECI,			0,				0,				0,				0,				IC_DECI,		IC_DECI		},
		{	ITOF,			0,				0,				0,				0,				IC_ITOF,		IC_ITOF		},
		{	FTOI,			0,				0,				0,				0,				IC_FTOI,		IC_FTOI		},
		{	INCX,			0,				0,				0,				0,				IC_INCX,		IC_INCX		},
		{	DECX,			0,				0,				0,				0,				IC_DECX,		IC_DECX		},
		{	GOTO,			0,				0,				0,				0,				IC_GOTO,		IC_GOTO		},
		{	IFTRUE,		0,				0,				0,				0,				IC_IFTRUE,	IC_IFTRUE	},
		{	IFFALSE,		0,				0,				0,				0,				IC_IFFALSE,	IC_IFFALSE	},
		{	GOTOI,		0,				0,				0,				0,				IC_GOTOI,	IC_GOTOI		},
		{	PUSHX,		0,				0,				0,				0,				IC_PUSHX,	IC_PUSHX		},
		{	POPXSP,		0,				0,				0,				0,				IC_POPXSP,	IC_POPXSP	},
		{	LOADSP,		0,				0,				0,				0,				IC_LOADSP,	IC_LOADSP	},
		{	SETSP,		0,				0,				0,				0,				IC_SETSP,	IC_SETSP		},
		{	SYSCALL,		0,				0,				0,				0,				IC_SYSCALL,	IC_SYSCALL	},
	};
	
	const int CODE_SCOPE_TABLE_SIZE = 6;
	const int codeScopeTable[][CODE_SCOPE_TABLE_SIZE] = {
	//		global			local
		{	IC_BLOAD,		IC_SPBLOAD		},
		{	IC_WLOAD,		IC_SPWLOAD		},
		{	IC_BSTORE,		IC_SPBSTORE		},
		{	IC_WSTORE,		IC_SPWSTORE		},
		{	IC_INC1,			IC_SPINC1		},
		{	IC_DEC1,			IC_SPDEC1		},
	};

	// Type sizes (in bytes) corresponding to the numeric type codes above;
	// string is a pointer (4 bytes).
	//
	const int SIZE[7] = { 4, 1, 1, 4, 4, 0, 0 };
	
	enum SizeMetric { BYTE, WORD };

	char* const INPUT_PATH_SWITCH	= "-path";
	char* const OUTPUT_FILE_EXT	= ".asm";
	char* const DEBUG_FILE_EXT		= ".dbg";

	extern std::map<std::string, std::string> inputArgs;
}

#endif
