/*
	Purpose	:	Source and AST parsers.
	Author	:	Alex Stojan
*/

header {
	#include "ANTLRExt.h"
}

header "post_include_hpp" {
	#include <iostream>

	#include "CodeGenerator.h"
	#include "Symbols.h"
	#include "Utilities.h"
	
	using namespace std;
	using namespace Compiler;
}

//#############################################################################

options {
	language="Cpp";
}

class SourceParser extends Parser;
options {
	k = 2;
	defaultErrorHandler = false;
	buildAST = true;
	ASTLabelType = "RefASTNodeExt";
}

tokens {
	// reserved words
	//
	PROGRAM = "program";
	TRUE = "true";
	FALSE = "false";
	NULLV = "null";
	RETURN = "return";
	LOOP = "loop";
	IF = "if";
	ELSE = "else";
	CONTINUE = "continue";
	BREAK = "break";
	WRITE = "write";
	WRITELN = "writeln";
	READ = "read";
	
	DECL;
	TYPE;
	ARRAY;
	SCOPE;
	BLOCK;
	CAST;
	DIM;
	CONST;
	CALL;
	FUNCTION;
	PARAMS;
	INIT;
	COND;
	ITER;
	LOOPHDR;
	ARGS;
	EMPTY;
	INC;
	DEC;
}


//#############################################################################
//#############################################################################
//#############################################################################


start:

	PROGRAM^ IDENT SEMI!

	// At this time only declarations are allowed to be global, but no
	// definitions. That's because no code is expected before the main block
	// and the first instruction is GOTO that jumps over all the functions into
	// the main block.
	
	(
			(declaration LPAREN ( paramList )? RPAREN LCURLY) => functionDef
			
		|	(declaration LPAREN ( paramTypeList )? RPAREN SEMI) => forward
		
		|	declaration SEMI!
	)*

	block
;
										
//--------------------------------------------------------------------------------------

forward:

	functionDecl
	
	{## = #([FORWARD, "FORWARD"], ##);}
;

//--------------------------------------------------------------------------------------

functionDecl:

	declaration LPAREN! ( paramTypeList )? RPAREN! SEMI!

	{## = #([FUNCTION, "FUNCTION"], ##);}
;

//--------------------------------------------------------------------------------------

paramTypeList:

	typeSpec ( COMMA! typeSpec )*

	{## = #([PARAMS, "PARAMS"], ##);}
;

//--------------------------------------------------------------------------------------

statement:

		BREAK^ SEMI!
		
	|	CONTINUE^ SEMI!
	
	|	block

	|	loop
	
	|	ifElse
	
	|	returnExpression SEMI!
	
	|	simpleStatement SEMI!

	|	( WRITE^ | WRITELN^ ) ( expression ( COMMA! expression )* )? SEMI!
	
	|	READ^ identifier ( COMMA! identifier )* SEMI!
;
	
//--------------------------------------------------------------------------------------

block:

	LCURLY! ( statement )* RCURLY!
	
	{## = #([BLOCK, "BLOCK"], ##);}
;

//--------------------------------------------------------------------------------------

functionDef:

	declaration LPAREN! ( paramList )? RPAREN! block

	{## = #([FUNCTION, "FUNCTION"], ##);}
;

//--------------------------------------------------------------------------------------

paramList:

	declaration ( COMMA! declaration )*

	{## = #([PARAMS, "PARAMS"], ##);}
;

//--------------------------------------------------------------------------------------

returnExpression:

	RETURN^ ( expression )?
;

//--------------------------------------------------------------------------------------

loop!:	// need to turn off tree construction for this rule to create AST manually

	LOOP
		LPAREN! 

		e1:loopInitializerList SEMI!
		e2:loopCondition SEMI!
		e3:loopIteration

		RPAREN! 
		
		( s:statement | SEMI! )
		
		// Here I put the iteration part AFTER the loop body, so the iteration
		// code will appear after the code for loop body.
		//
		{
			if (#s == 0) #s = #[EMPTY, "EMPTY"];
			## = #([LOOPHDR, "loop"], #e1, #e2, #s, #e3);
		}
;

//--------------------------------------------------------------------------------------

loopInitializerList:

	( simpleStatement ( COMMA! simpleStatement )* | )

	{## = #([INIT, "INIT"], ##);}
;

//--------------------------------------------------------------------------------------

loopCondition:

	( expression | )

	{## = #([COND, "COND"], ##);}
;

//--------------------------------------------------------------------------------------

loopIteration:

	( simpleStatement ( COMMA! simpleStatement )* | )

	{## = #([ITER, "ITER"], ##);}
;

//--------------------------------------------------------------------------------------

ifElse:
	IF^ LPAREN! expression RPAREN! statement

	( 
		// CONFLICT: "Dangling-else" problem - ANTLR generates correct code by matching ASAP
		//
		options {
			warnWhenFollowAmbig = false;
		}:

		ELSE! statement
	)?
;

//--------------------------------------------------------------------------------------

declaration:

	type:typeSpec id:IDENT^

	{## = #(#([DECL, "DECL"], ##), type, [IDENT, id->getText()]);}
;

//---------------------------------------------------------------------------------------------

definition:
	declaration	( ASSIGN^ expression )?
;

//---------------------------------------------------------------------------------------------

arraySpec:

	( LBRACKET! NUMERIC RBRACKET! )+
	
	{## = #([ARRAY, "ARRAY"], ##);}
;

//---------------------------------------------------------------------------------------------

typeSpec:

	IDENT ( REF )* ( arraySpec )?
	
	{## = #([TYPE, "TYPE"], ##);}
;

//---------------------------------------------------------------------------------------------
//***** SIMPLE EXPRESSIONS *****
//---------------------------------------------------------------------------------------------

simpleStatement:

		( typeSpec IDENT ) => definition
		
	|	incDec
	
	|	assign
;
										
//--------------------------------------------------------------------------------------

incDec:

		INCR! specialUnaryExpr

		{## = #([INC, "INC"], ##);}

	|	DECR! specialUnaryExpr
	
		{## = #([DEC, "DEC"], ##);} 
;	

//--------------------------------------------------------------------------------------

assign:

	expression ( ASSIGN^ expression )?
;

//--------------------------------------------------------------------------------------

expression:
	orExpr
;

//--------------------------------------------------------------------------------------

orExpr:
	andExpr ( OR^ andExpr )*
;

//--------------------------------------------------------------------------------------

andExpr:
	logPrimary ( AND^ logPrimary )*
;
	
//--------------------------------------------------------------------------------------

logPrimary:
	relExpr ( ( NEQ^ | EQ^ ) relExpr )*
;
	
//--------------------------------------------------------------------------------------

relExpr:
	additiveExpr ( ( TK_LT^ | GT^ | LE^	| GE^ ) additiveExpr )*
;

//--------------------------------------------------------------------------------------

additiveExpr:
	multiplicativeExpr ( ( PLUS^ | MINUS^ ) multiplicativeExpr )*
;

//--------------------------------------------------------------------------------------

multiplicativeExpr:
	unaryExpr ( ( STAR^ | DIVIDE^ | MODULO^ ) unaryExpr )*
;

//--------------------------------------------------------------------------------------

unaryExpr:

		( MINUS^ | NOT^ ) unaryExpr

	|	specialUnaryExpr
;
										
//--------------------------------------------------------------------------------------

specialUnaryExpr:

		( LPAREN typeSpec RPAREN specialUnaryExpr ) => LPAREN! typeSpec RPAREN! specialUnaryExpr

		{## = #([CAST, "CAST"], ##);}

	|	ADDR^ postfixExpr

	|	REF^ specialUnaryExpr
	
	|	postfixExpr
;

//--------------------------------------------------------------------------------------

postfixExpr:

	primaryExpr	( LBRACKET^ expression RBRACKET! ( LBRACKET! expression RBRACKET! )* )?
;

//--------------------------------------------------------------------------------------

primaryExpr:

		LPAREN! expression RPAREN!
		
	|	( IDENT LPAREN ) => call
		
	|	identifier

	|	constant
;
	
//--------------------------------------------------------------------------------------

call:

	id:IDENT LPAREN! ( args )? RPAREN!
	
	{## = #([CALL, "CALL"], ##);}
;

//--------------------------------------------------------------------------------------

args:

	expression ( COMMA! expression )*
	
	{## = #([ARGS, "ARGS"], ##);}
;

//--------------------------------------------------------------------------------------

constant
{
	string value;
}:

		n:NUMERIC

		{ 
			value = n->getText();
			size_t p = value.find(".");
			
			## = #(#([CONST, "CONST"], ##), #[NUMERIC, value]);
		}

	|	c:CHAR

		{
			value = c->getText();
			value = value.substr(1, value.size() - 2);

			## = #(#([CONST, "CONST"], ##), #[CHAR, value]);
		}

	|	t:TRUE

		{
			value = t->getText();

			## = #(#([CONST, "CONST"], ##), #[TRUE, value]);
		}

	|	f:FALSE

		{
			value = f->getText();

			## = #(#([CONST, "CONST"], ##), #[FALSE, value]);
		}

	|	z:NULLV
		
		{
			value = z->getText();


			## = #(#([CONST, "CONST"], ##), #[NULLV, value]);
		}
		
	|	s:STRING
	
		{
			value = s->getText();

			## = #(#([CONST, "CONST"], ##), #[STRING, value]);
		}
;

//--------------------------------------------------------------------------------------

identifier:
	id:IDENT
	
	{## = #([IDENT, "IDENT"], ##);}
;


//#############################################################################
//#############################################################################
//#############################################################################


class SourceLexer extends Lexer;
options {
	k = 2;
	defaultErrorHandler = false;
}

WS	:	(	' '
		|	'\t'
		|	'\f'
		// handle newlines
		|	(	"\r\n"  // Evil DOS
			|	'\r'    // Macintosh
			|	'\n'    // Unix (the right way)
			)
			{ newline(); }
		)
		{ _ttype = ANTLR_USE_NAMESPACE(antlr)Token::SKIP; }
	;

LCURLY		:	'{';
RCURLY		:	'}';
LPAREN		:	'(';
RPAREN		:	')';
STAR			:	'*';
DIVIDE		:	'/';
PLUS			:	'+';
MINUS			:	'-';
MODULO		:	'%';
SEMI			:	';';
LBRACKET		:	'[';
RBRACKET		:	']';
COMMA			:	',';
DOT			:	'.';
ARROW			:	"->";
TK_LT			:	'<';
GT				:	'>';
LE				:	"<=";
GE				:	">=";
EQ				:	'=';
NEQ			:	"#";
ASSIGN		:	":=";
OR				:	'|';
AND			:	'&';
NOT			:	'~';
REF			:	'^';
ADDR			:	'@';
DIRECTIVE	:	'$';
INCR			:	"++";
DECR			:	"--";

INS_START	:	"${";
INS_END		:	"$}";

NUMERIC		:	(DIGIT)+ ( DOT (DIGIT)+ )?
				;

IDENT			:	(LETTER) (LETTER | DIGIT)*
				;
							
STRING		:	'"' ( ~('\\'|'"') )* '"'
				;

CHAR			:	'\'' (~'\'') '\''
				;

COMMENT		:	"//" (~('\n'|'\r'))* ('\n'|'\r'('\n')?)
					{
						$setType(ANTLR_USE_NAMESPACE(antlr)Token::SKIP);
						newline();
					}
				;

COMMENT_2	:	"/*"

					( 
						options { generateAmbigWarnings=false; }
						:	{ LA(2) != '/' }? '*'
						|	'\r' '\n'	{ newline(); }
						|	'\r'			{ newline(); }
						|	'\n'			{ newline(); }
					  	|   ~('*' | '\n' | '\r')
					)*

					"*/"
		
					{$setType(ANTLR_USE_NAMESPACE(antlr)Token::SKIP);}
				;

protected
DIGIT			:	('0'..'9');

protected
LETTER		:	('a'..'z' | 'A'..'Z' | '_');


//#############################################################################
//#############################################################################
//#############################################################################


{
	#include <stack>
	
	typedef vector<int>					AddrVec;
	typedef pair<AddrVec, AddrVec>	BreakContinuePair;
	typedef vector<antlr::RefAST>		TopLevelCallVector;
	
	Generator						gen;
	stack<BreakContinuePair>	loopStack;
	Scope								scope;
}

class ASTParser extends TreeParser;
options {
	defaultErrorHandler = false;
 	ASTLabelType = "RefASTNodeExt";
}

{
	// For each block we need to know whether a returned value from
	// a function is used or not, so we can adjust the expression
	// stack pointer accordingly upon return. Here I just find each
	// such top-level call and set the indicator, so when we get to
	// the call node we can check if it's a top-level call.
	//
	void findTopLevelCalls(RefASTNodeExt startNode, bool skipFirstChild = false) {
		RefASTNodeExt node = RefASTNodeExt(startNode->getFirstChild());
		
		if (skipFirstChild) {
			node = node->getNextSibling();
		}
		
		while (node) {
			if (node->getType() == CALL) {
				node->isTopLevelCall = true;
			}
			
			node = node->getNextSibling();
		}
	}
	
	void ERROR_IF(bool cond, int line, string msg) {
		if (cond) {
			throw CompilerException(line, msg);
		}
	}
}
//--------------------------------------------------------------------------------------

start
{
	Symbol* sym = 0;
}:

	#(p:PROGRAM programName:IDENT 
	
			{
				// we need this GOTO to jump over all functions into main block
				gen.genAsm(SymbolicInstr(Compiler::GOTO, Constant(0)));

				scope.enter(0);
			}
	
		( functionDef | forward | sym = declaration )*
	
			{
				gen.backpatchAt(0, Constant(gen.getCodeSize()));
			}
		
		block
	)
	
		{
			gen.patchFunctionCalls(scope.top());
			scope.exit();
			gen.genAsm(SymbolicInstr(Compiler::NOP));
			gen.generateAssembly(programName->getText(), scope.allSymbolTables);
			
			cout << "Generated " << programName->getText() << OUTPUT_FILE_EXT << endl;
		}
;
										
//--------------------------------------------------------------------------------------

forward
{
	Symbol* fn = 0;
	SymbolTable* localST = 0;
	TypeAttr type;
	Symbol::ParamVector params;
}:

	#(FORWARD
		#(f:FUNCTION
			fn = declaration[false, true]
		
				{
					fn->isFunction = true;
					fn->isForwardedFunction = true;
				}

			(#(PARAMS 
					(type = typeSpec

						{
							fn->fwdParamTypes.push_back(type);
						}
					)+
					
						{
							bool result = scope.getTopLevel()->addFunction(fn);
							
							ERROR_IF(!result, f->getLine(), "Illegal forward declaration");
						}
				) 
			)?
		)
	)
;

//--------------------------------------------------------------------------------------

statement:

		continueStmt

	|	breakStmt

	|	block

	|	loop

	|	ifElse

	|	returnExpression

	|	simpleStatement
	
	|	inputOutputStmt
;
	
//--------------------------------------------------------------------------------------

functionDef
{
	Symbol* fn = 0;
	SymbolTable* localST = 0;
}:

	#(f:FUNCTION
		fn = declaration[false, true]

			{
				fn->index = gen.getCodeSize();
				fn->isFunction = true;
				fn->localAddressCounter = fn->type.getSize(WORD, TypeAttr::ARRAY_AS_POINTER) * WORD_SIZE;
				
				gen.currentFunction = fn;
				
				scope.resetSpace();				// we need to know the space size for local data
				localST = scope.enter(0);		// create the local symbol table
				localST->function = fn;
			}
	
		( paramList[fn] )?
		
			{
				bool result = scope.getTopLevel()->addFunction(fn);

				ERROR_IF(!result, f->getLine(), "Illegal function definition");
			}

		block[false, fn]		// scope already created for parameters
	)

		{
			fn->localSpace = scope.exit();
			fn->isFullFunctionDef = true;

			if (fn->type.getSize() == 0) {	// generate return code for void function
				// ### load the return address
				gen.genAsm(SymbolicInstr(Compiler::LOAD, 
													Constant((fn->localSpace + fn->type.getSize(WORD, TypeAttr::ARRAY_AS_POINTER)) * WORD_SIZE),
													INT_CODE, 
													SymbolicInstr::OPERAND));

				// ### jump to the return address
				gen.genAsm(SymbolicInstr(Compiler::GOTOI));

				gen.currentFunction = 0;
			}

			// ### backpatch all instructions that need the function space info
			
			gen.backpatch(fn, AFTER_FN_BODY, fn->localSpace);
			gen.currentFunction = 0;
			gen.hasReturnStatement = false;
		}
;

//--------------------------------------------------------------------------------------

paramList[Symbol* fn]
{
	Symbol* sym = 0;
	Symbol::ParamVector params;
}:

	#(PARAMS 
		(
			sym = declaration[true]
		
				{
					sym->isLocal = true;
					sym->isParameter = true;
					params.push_back(sym);
				}
		)+
	)
	
		{
			fn->setParams(params);
		}
;

//--------------------------------------------------------------------------------------

inputOutputStmt
{
	TypeAttr type;
	Symbol* sym = 0;
	int args = 0;
	int sysCode = 0;
}:

		(
				#(WRITE { sysCode = SYSCODE_WRITE; }

					( 
						type = expression 

							{
								gen.genAsm(SymbolicInstr(Compiler::ICLOAD, Constant(type.getTypeNameCode())));		// arg type
								++args;
							}
					)*
				)
			
			|	#(WRITELN { sysCode = SYSCODE_WRITELN; }

					( 
						type = expression 

							{
								gen.genAsm(SymbolicInstr(Compiler::ICLOAD, Constant(type.getTypeNameCode())));		// arg type
								++args;
							}
					)*
				)
		)

			{
				gen.genAsm(SymbolicInstr(Compiler::SYSCALL, Constant(sysCode), Constant(args)));
			}

	|	#(READ
				( 
					sym = identifier[false]

						{
							gen.genAddressOf(sym);
							gen.genAsm(SymbolicInstr(Compiler::ICLOAD, Constant(sym->type.getTypeNameCode())));		// arg type
							++args;
						}
				)+
		)
		
			{
				gen.genAsm(SymbolicInstr(Compiler::SYSCALL, Constant(SYSCODE_READ), Constant(args)));
			}
;

//--------------------------------------------------------------------------------------

continueStmt:

	n:CONTINUE

		{
			ERROR_IF(loopStack.size() == 0, n->getLine(), "The 'continue' statement can be used only inside a loop body");

			BreakContinuePair& bcPair = loopStack.top();
			AddrVec* vec = &(bcPair.second);
			vec->push_back(gen.getCodeSize());
			gen.genAsm(SymbolicInstr(Compiler::GOTO, Constant(0)));
		}
;

//--------------------------------------------------------------------------------------

breakStmt:

	n:BREAK
	
		{
			ERROR_IF(loopStack.size() == 0, n->getLine(), "The 'break' statement can be used only inside a loop body");

			BreakContinuePair& bcPair = loopStack.top();
			AddrVec* vec = &(bcPair.first);
			vec->push_back(gen.getCodeSize());
			gen.genAsm(SymbolicInstr(Compiler::GOTO, Constant(0)));
		}
;

//--------------------------------------------------------------------------------------

block[bool createScope = true, Symbol* function = 0]:

	#(b:BLOCK 

		{
			// This code verifies that a non-void function returns a value. It only
			// scans the outermost block of the function; it doesn't check all the
			// control paths to ensure that all of them return a value (which would
			// be the right way to do this). Therefore, the compiler requires all
			// non-void functions to have a return statement as the last statement
			// of the outermost block. It also verifies that there are no 
			// statements following a return statement.

			bool hasReturnStatement = false;
			antlr::RefAST node = b->getFirstChild();
			while (node) {
				if (node->getType() == RETURN) {
					hasReturnStatement = true;

					// the return statement must be the last statement in the block
					ERROR_IF(node->getNextSibling() != 0,
								b->getLine(), "The return statement is not the last statement in the block");
					
					break;
				}
				
				node = node->getNextSibling();
			}
				
			// make sure that non-void functions have a return statement
			ERROR_IF(function && function->type.getSize() > 0 && !hasReturnStatement,
						b->getLine(), "The outermost block must have 'return' as the last statement");
		}

		( 
				{
					// create the symbol table for the block
					if (createScope) scope.enter(gen.currentFunction);

					// find all top-level function calls (if any)
					findTopLevelCalls(b);
				}
			
			statementList 
			
				{	// the end of block - go back to the outer symbol table (if any)
					if (createScope) scope.exit();
				}
		)?
	)
;

//--------------------------------------------------------------------------------------

statementList:

	( statement )+
;

//--------------------------------------------------------------------------------------

returnExpression
{
	TypeAttr t;
	bool hasExpression = false;
}:

	#(r:RETURN 
		(
			t = expression

				{
					hasExpression = true;
				}
		)?
	)

			{
				ERROR_IF(! gen.currentFunction, r->getLine(), "The 'return' statement cannot be used outside of a function body");
			
				ERROR_IF(gen.currentFunction->type.getSize() > 0 && !hasExpression,
							r->getLine(), "Non-void functions must return a value");
				
				ERROR_IF(gen.currentFunction->type.getSize() == 0 && hasExpression,
							r->getLine(), "Void functions cannot return a value");

				gen.hasReturnStatement = true;

				struct SetVal: public IBackpatch {
					SetVal(int p, int v, Symbol* sym, BackpatchTime when): IBackpatch(p, v, sym, when) {}
					int value(int x) const { return initial + x * WORD_SIZE; }
				};
			
				// ### load the return address
				gen.bp.push_back(new SetVal(gen.getCodeSize(), 
														gen.currentFunction->type.getSize(WORD, TypeAttr::ARRAY_AS_POINTER) * WORD_SIZE,
														gen.currentFunction, 
														AFTER_FN_BODY));
				gen.genAsm(SymbolicInstr(Compiler::LOAD, Constant(0), INT_CODE, SymbolicInstr::OPERAND));
				
				// ### jump to the return address
				gen.genAsm(SymbolicInstr(Compiler::GOTOI));
			}
;

//--------------------------------------------------------------------------------------

loop
{
	TypeAttr t;
	int loopStart = 0;
	int backpatchInstrLoc = 0;
	bool condExists = false;
	bool statementExists = false;
	bool iterationExists = false;
	int loopID = gen.getCodeSize();
	int loopIterAddr = 0;
}:

	#(l:LOOPHDR
	
			{
				findTopLevelCalls(l);

				// Add the vectors to the loop stack that will serve as placeholder
				// for BREAK/CONTINUE addresses. These addresses will then be
				// back-patched with the address of the first instruction following
				// the loop (in case of BREAK) or the beginning of the iteration
				// part (in case of CONTINUE).
				//
				loopStack.push(BreakContinuePair());

				// Define the scope for the loop - any declarations in a loop
				// header are visible only inside the loop header and body.
				//
				scope.enter(gen.currentFunction);
			}

		#(init:INIT

				{
					findTopLevelCalls(init);
				}

			( simpleStatement )*
		) 
		
			{
				loopStart = gen.getCodeSize();
			}

		#(COND 
			( 
				t = expression 

					{
						condExists = true;
						backpatchInstrLoc = gen.getCodeSize();
						gen.genAsm(SymbolicInstr(Compiler::IFFALSE, Constant(0)));
					}
			)*
		) 
		
		(
				statement	{ statementExists = true; }
			|
				EMPTY 
		)
		
		#(iter:ITER 

				{
					findTopLevelCalls(iter);
				}

			( 
					{ 
						iterationExists = true; 

						// might be needed to backpatch CONTINUE statement
						//
						loopIterAddr = gen.getCodeSize();
					} 

				simpleStatement 
			)*
		)
	)

		{
			if (statementExists || condExists || iterationExists) {
				gen.genAsm(SymbolicInstr(Compiler::GOTO, Constant(loopStart)));
			}

			if (condExists) {
				gen.backpatchAt(backpatchInstrLoc, Constant(gen.getCodeSize()));
			}

			// get the loop's stack and backpatch BREAK/CONTINUE (if needed)
			
			BreakContinuePair& bcPair = loopStack.top();
			AddrVec breakVec = bcPair.first;
			AddrVec contVec = bcPair.second;
			
			int backpatchLoc = 0;
			
			// ... check for BREAK
			//
			for (size_t i = 0;  i < breakVec.size();  ++i) {
				backpatchLoc = breakVec[i];
				gen.backpatchAt(backpatchLoc, Constant(gen.getCodeSize()));
			}
			
			// ... check for CONTINUE
			//
			for (size_t i = 0;  i < contVec.size();  ++i) {
				backpatchLoc = contVec[i];
				gen.backpatchAt(backpatchLoc, Constant(loopIterAddr));
			}

			loopStack.pop();
			scope.exit();
		}
;

//--------------------------------------------------------------------------------------

ifElse
{
	TypeAttr t;
	int backpatchCond = 0;
	int backpatchThen = 0;
	bool elseExists = false;
}:

	#(n:IF 

			{
				findTopLevelCalls(n, true);
			}

		t = expression 

			{
				backpatchCond = gen.getCodeSize();
				gen.genAsm(SymbolicInstr(Compiler::IFFALSE, Constant(0)));
			}

		statement 
		
			{
				backpatchThen = gen.getCodeSize();
				gen.genAsm(SymbolicInstr(Compiler::GOTO, Constant(0)));
			}

		( 
				{
					elseExists = true;
					gen.backpatchAt(backpatchCond, Constant(gen.getCodeSize()));
				}

			statement 

				{
					gen.backpatchAt(backpatchThen, Constant(gen.getCodeSize()));
				}
		)?

			{
				if (! elseExists) {
					gen.removeLastInstr();		// no ELSE, so we don't need the GOTO to jump over it
					gen.backpatchAt(backpatchCond, Constant(gen.getCodeSize()));
				}
			}
	)
;

//--------------------------------------------------------------------------------------

declaration[bool isParameter = false, bool createSymbol = false]
returns [Symbol* sym = 0]
{
	TypeAttr type;
}:

	#(node:DECL type = typeSpec id:IDENT)
	
		{
			if (createSymbol) {
				// this will be deleted by the SymbolTable destructor
				sym = new Symbol(id->getText(), scope.top()->getID(), type);
			}
			else {
				sym = scope.top()->addSymbol(id->getText(), type, isParameter);
				
				ERROR_IF(sym == 0, node->getLine(), "Duplicate identifier");
				
				sym->isLocal = gen.currentFunction != 0;
			}
		}
;

//---------------------------------------------------------------------------------------------

arrayTypeSpec
returns [vector<int> dim]
{
	Constant cns;
}:

	#(ARRAY 
		( 
			n:NUMERIC
			
				{
					Constant cns = getNumericConstant(n->getText());
					
					ERROR_IF(cns.type != TypeAttr(INT),
								n->getLine(), "Array dimensions must be specified using integer constants");
					
					dim.push_back(cns.intVal);
				}
		)+ 
	)
;

//---------------------------------------------------------------------------------------------

typeSpec
returns [TypeAttr t]
{
	vector<int> dim;
}:

	#(TYPE typeName:IDENT
		( 
			REF 
			
				{
					t.setAsPointer();
					t.increaseRefLevel();
				}
		)*
		
		( dim = arrayTypeSpec )?
	)
	
		{
			bool typeExists = t.setTypeName(typeName->getText(), dim);

			ERROR_IF(!typeExists, typeName->getLine(), "Undefined type");
		}
;

//---------------------------------------------------------------------------------------------

simpleStatement
{
	TypeAttr t;
	Symbol* sym;
}:

		sym = declaration 
		
	|	incDec
		
	|	assignment
	
	|	t = expression
;

//--------------------------------------------------------------------------------------

incDec
{
	TypeAttr type;
	Symbol* sym = 0;
}:

		#(incNode:INC 
			( 
					(identifier) => sym = identifier[false]
					
						{
							ERROR_IF(sym->type != TypeAttr(INT) && sym->type != TypeAttr(Compiler::CHAR),
										incNode->getLine(), "Operator '++' is defined only for integer types");
							
							gen.genAsm(SymbolicInstr(Compiler::INC1, sym));
						}
					
				|	(reference) => type = reference[true, true]

						{
							ERROR_IF(type != TypeAttr(INT) && type != TypeAttr(Compiler::CHAR),
										incNode->getLine(), "Operator '++' is defined only for integer types");

							gen.genAsm(SymbolicInstr(Compiler::ICLOAD, Constant(1)));
							gen.genAsm(SymbolicInstr(Compiler::INCI));
						}

				|	type = expression 

						{
							ERROR_IF(type != TypeAttr(INT) && type != TypeAttr(Compiler::CHAR),
										incNode->getLine(), "Operator '++' is defined only for integer types");

							gen.genAsm(SymbolicInstr(Compiler::ICLOAD, Constant(1)));
							gen.genAsm(SymbolicInstr(Compiler::INCI));
						}
			)
		)
		
	|	#(decNode:DEC 
			( 
					(identifier) => sym = identifier[false]
					
						{
							ERROR_IF(sym->type != TypeAttr(INT) && sym->type != TypeAttr(Compiler::CHAR),
										decNode->getLine(), "Operator '--' is defined only for integer types");

							gen.genAsm(SymbolicInstr(Compiler::DEC1, sym));
						}

				|	(reference) => type = reference[true, true]

						{
							ERROR_IF(type != TypeAttr(INT) && type != TypeAttr(Compiler::CHAR),
										decNode->getLine(), "Operator '--' is defined only for integer types");

							gen.genAsm(SymbolicInstr(Compiler::ICLOAD, Constant(1)));
							gen.genAsm(SymbolicInstr(Compiler::DECI));
						}

				|	type = expression 

						{
							ERROR_IF(type != TypeAttr(INT) && type != TypeAttr(Compiler::CHAR),
										decNode->getLine(), "Operator '--' is defined only for integer types");

							gen.genAsm(SymbolicInstr(Compiler::ICLOAD, Constant(1)));
							gen.genAsm(SymbolicInstr(Compiler::DECI));
						}
			)
		)

;

//--------------------------------------------------------------------------------------

assignment
{
	Symbol* sym;
	TypeAttr lt, rt;
}:

	#(node:ASSIGN
		(
				sym = declaration rt = expression			// definition

					{
						ERROR_IF(! rt.isAssignableTo(sym->type), node->getLine(), "Illegal definition");

						gen.genAsm(SymbolicInstr(Compiler::STORE, sym, sym->type.getStorageClass()));
					}
					
			|	lt = arrayAccess rt = expression				// assignment to an array index
				
					{
						ERROR_IF(! rt.isAssignableTo(lt), node->getLine(), "Illegal assignment to array index");

						gen.genAsm(SymbolicInstr(Compiler::STOREI, lt.getStorageClass(lt.isPointer())));
					}
					
			|	lt = reference[true] rt = expression		// assignment to a pointer
				
					{
						ERROR_IF(! rt.isAssignableTo(lt), node->getLine(), "Illegal assignment to pointer");

						gen.genAsm(SymbolicInstr(Compiler::STOREI0, lt.getStorageClass()));
					}
					
			|	sym = identifier[false] rt = expression	// assignment to an identifier

					{
						ERROR_IF(! rt.isAssignableTo(sym->type), node->getLine(), "Illegal assignment to identifier");

						gen.genAsm(SymbolicInstr(Compiler::STORE, sym, sym->type.getStorageClass()));
					}
		)
	)
;

//--------------------------------------------------------------------------------------

expression
returns [TypeAttr type]
{
	TypeAttr rt;
	Symbol* sym = 0;
	int loc = 0;
}:

		#(orNode:OR type = expression rt = expression)
		
			{ 
				ERROR_IF(type != TypeAttr(BOOL) || rt != TypeAttr(BOOL),
							orNode->getLine(), "Cannot apply the '|' operator to a non-boolean value");

				gen.genAsm(SymbolicInstr(Compiler::OR)); 
				type = TypeAttr(BOOL);
			}

	|	#(andNode:AND type = expression rt = expression)
	
			{ 
				ERROR_IF(type != TypeAttr(BOOL) || rt != TypeAttr(BOOL),
							andNode->getLine(), "Cannot apply the '&' operator to a non-boolean value");

				gen.genAsm(SymbolicInstr(Compiler::AND)); 
				type = TypeAttr(BOOL);
			}

	|	#(NEQ 
			type = expression 
			
				{
					loc = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::NOP));		// reserve for possible int-to-float conversion
				}
				
			rt = expression)
	
			{ 
				gen.generateArithmeticInstr(Compiler::NEQ, type, rt, loc);
				type = TypeAttr(BOOL);
			}

	|	#(eqNode:EQ 
			type = expression 
			
				{
					loc = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::NOP));		// reserve for possible int-to-float conversion
				}
				
			rt = expression)
	
			{ 
				gen.generateArithmeticInstr(Compiler::EQ, type, rt, loc);
				type = TypeAttr(BOOL);
			}

	|	#(ltNode:TK_LT 
			type = expression 
			
				{
					loc = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::NOP));		// reserve for possible int-to-float conversion
				}
				
			rt = expression)
	
			{ 
				ERROR_IF(! type.isNumericOrChar() || ! rt.isNumericOrChar(),
							ltNode->getLine(), "Cannot apply the '<' operator to a non-numeric value");

				gen.generateArithmeticInstr(Compiler::LT, type, rt, loc);
				type = TypeAttr(BOOL);
			}

	|	#(gtNode:GT 
			type = expression 
			
				{
					loc = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::NOP));		// reserve for possible int-to-float conversion
				}
				
			rt = expression)
	
			{ 
				ERROR_IF(! type.isNumericOrChar() || ! rt.isNumericOrChar(),
							gtNode->getLine(), "Cannot apply the '>' operator to a non-numeric value");

				gen.generateArithmeticInstr(Compiler::GT, type, rt, loc);
				type = TypeAttr(BOOL);
			}

	|	#(leNode:LE 
			type = expression 
			
				{
					loc = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::NOP));		// reserve for possible int-to-float conversion
				}
				
			rt = expression)
	
			{ 
				ERROR_IF(! type.isNumericOrChar() || ! rt.isNumericOrChar(),
							leNode->getLine(), "Cannot apply the '<=' operator to a non-numeric value");

				gen.generateArithmeticInstr(Compiler::LE, type, rt, loc);
				type = TypeAttr(BOOL);
			}

	|	#(geNode:GE 
			type = expression 
			
				{
					loc = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::NOP));		// reserve for possible int-to-float conversion
				}
				
			rt = expression)
	
			{ 
				ERROR_IF(! type.isNumericOrChar() || ! rt.isNumericOrChar(),
							geNode->getLine(), "Cannot apply the '>=' operator to a non-numeric value");

				gen.generateArithmeticInstr(Compiler::GE, type, rt, loc);
				type = TypeAttr(BOOL);
			}

	|	#(plusNode:PLUS 
			type = expression 
			
				{
					loc = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::NOP));		// reserve for possible int-to-float conversion
				}
				
			rt = expression)
	
			{ 
				ERROR_IF(! type.isNumericOrChar() || ! rt.isNumericOrChar(),
							plusNode->getLine(), "Cannot apply the '+' operator to a non-numeric value");

				type = gen.generateArithmeticInstr(Compiler::ADD, type, rt, loc);
			}
	
	|	(#(MINUS expression expression)) => 
			#(minNode:MINUS 
				type = expression 
			
					{
						loc = gen.getCodeSize();
						gen.genAsm(SymbolicInstr(Compiler::NOP));		// reserve for possible int-to-float conversion
					}
				
				rt = expression)

			{ 
				ERROR_IF(! type.isNumericOrChar() || ! rt.isNumericOrChar(),
							minNode->getLine(), "Cannot apply the '-' operator to a non-numeric value");

				type = gen.generateArithmeticInstr(Compiler::SUB, type, rt, loc);
			}
	
	|	#(mulNode:STAR 
			type = expression 
			
				{
					loc = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::NOP));		// reserve for possible int-to-float conversion
				}
				
			rt = expression)
	
			{ 
				ERROR_IF(! type.isNumericOrChar() || ! rt.isNumericOrChar(),
							mulNode->getLine(), "Cannot apply the '*' operator to a non-numeric value");

				type = gen.generateArithmeticInstr(Compiler::MUL, type, rt, loc);
			}

	|	#(divNode:DIVIDE 
			type = expression 
			
				{
					loc = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::NOP));		// reserve for possible int-to-float conversion
				}
				
			rt = expression)
	
			{
				ERROR_IF(! type.isNumericOrChar() || ! rt.isNumericOrChar(),
							divNode->getLine(), "Cannot apply the '/' operator to a non-numeric value");

				type = gen.generateArithmeticInstr(Compiler::DIV, type, rt, loc);
			}

	|	#(modNode:MODULO type = expression rt = expression)
	
			{
				ERROR_IF(type != TypeAttr(INT) || rt != TypeAttr(INT),
							modNode->getLine(), "Cannot apply the '%' operator to a non-integer value");

				gen.genAsm(SymbolicInstr(Compiler::MOD));
			}

	|	#(invNode:MINUS type = expression)
	
			{ 
				ERROR_IF(! type.isNumericOrChar(),
							invNode->getLine(), "Cannot apply unary '-' operator to a non-numeric value");

				gen.genAsm(SymbolicInstr(Compiler::NEG, type.getTypeNameCode())); 
			}

	|	#(negNode:NOT type = expression)
	
			{
				ERROR_IF(type != TypeAttr(BOOL), negNode->getLine(), "Cannot apply the '~' operator to a non-boolean value");

				gen.genAsm(SymbolicInstr(Compiler::NOT));
			}

	|	#(ADDR 
			(	sym = identifier[false] 

					{
						gen.genAddressOf(sym);

						type = sym->type;
						type.clearArrayAttributes();		// in case the identifier was an array
						type.increaseRefLevel();
					}

				|	type = arrayAccess

						{
							// multiply by the object size to get to the right position
							//
							if (getTypeSizeFromTypeName(type.getName()) >= WORD_SIZE) {
								gen.genAsm(SymbolicInstr(Compiler::ICLOAD, Constant(getTypeSizeFromTypeName(type.getName()))));
								gen.genAsm(SymbolicInstr(Compiler::MUL, type.getTypeNameCode()));
							}
							
							gen.genAsm(SymbolicInstr(Compiler::ADD, type.getTypeNameCode()));
							
							// since we took the address of an array index we got a pointer type
							
							type.clearArrayAttributes();		// in case the identifier was an array
							type.increaseRefLevel();
						}
			)
		)

	|	type = reference

	|	#(CAST type = typeSpec rt = expression)
	
			{
				if (rt == TypeAttr(INT) && type == TypeAttr(FLOAT)) {
					gen.genAsm(SymbolicInstr(Compiler::ITOF));
				}
				else {
					if (rt == TypeAttr(FLOAT) && type == TypeAttr(INT)) {
						gen.genAsm(SymbolicInstr(Compiler::FTOI));
					}
				}
			}
	
	|	type = functionCall

	|	type = primaryExpr
;
										
//--------------------------------------------------------------------------------------

// The general structure of a call frame is this:
//
//		+--------------------+
//		|	Previous SP			|
//		|	Return Address		|
//		|	Locals				|	(optional)
//		|	Parameters			|	(optional)
//		|	Result				|	(optional)
//		+--------------------+
//		|			...			|
//		|	<previous frame>	|
//		|			...			|
//
functionCall
returns [TypeAttr type]
{
	Symbol* sym = 0;
	bool useReturnValue = true;
	FunctionCall* call = new FunctionCall;
	vector<TypeAttr> argTypes;
}:

	#(fc:CALL 
		procName:IDENT 
		
			{
				//sym = findSymbol(procName->getText(), scope.stStack);
				sym = findFunction(procName->getText(), scope);
				
				ERROR_IF(! sym, procName->getLine(), "Undefined function");

				if (fc->isTopLevelCall) {
					useReturnValue = false;
				}
				else {	// someone is using the return value - make sure the function is not void
					ERROR_IF(sym->type.getSize() == 0, fc->getLine(), "Function does not return a value");
				}

				call->name = sym->name;

				// ### save the expression stack pointer
				gen.genAsm(SymbolicInstr(Compiler::PUSHX));

				// The return value (if any) is stored at the bottom of the frame
				// because when the function returns the expression stack pointer
				// will be restored such that to include this return value into
				// the previous expression.

				// ### make room for the return value (if there is one)
				if (sym->type.getSize() > 0) {
					call->resultSpace = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::INCX, Constant(0)));
				}
			}

		// If there are any arguments they will be loaded into area right above
		// the area for the return value. If there is no return value the
		// arguments will be loaded at the bottom of the frame.
			
		( argTypes = args )?

			{
				ERROR_IF(!validateFunctionCall(sym, &argTypes), procName->getLine(), "Invalid function call");

				// ### make room for local non-parameter data

				// With forward calls we don't know what the local space is,
				// otherwise we do, so we can save some time.
				//
				if (sym->isFullFunctionDef) {
					// if the function doesn't have any local non-parameter data then don't generate INCX instruction
					if (sym->localSpace - sym->paramsSpace > 0) {
						call->nonParamSpace = gen.getCodeSize();
						gen.genAsm(SymbolicInstr(Compiler::INCX, Constant(0)));
					}
				}
				else {	// we don't have a full def (it's a forward call) so we can't assume anything
					call->nonParamSpace = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::INCX, Constant(0)));
				}

				// ### load the return address
				gen.genAsm(SymbolicInstr(Compiler::ICLOAD, Constant(gen.getCodeSize() + 4)));

				// ### save the current SP
				gen.genAsm(SymbolicInstr(Compiler::LOADSP));

				// ### set the new SP
				gen.genAsm(SymbolicInstr(Compiler::POPXSP));

				// ### jump to the function code
				call->targetAddr = gen.getCodeSize();
				gen.genAsm(SymbolicInstr(Compiler::GOTO, Constant(0)));

				// the following instructions are executed after the function returns

				// ### adjust the expression stack pointer

				// If the function's return value is not used we need to adjust
				// the expr. stack pointer accordingly.
				//
				call->restoreX = gen.getCodeSize();
				gen.genAsm(SymbolicInstr(Compiler::DECX, Constant(0)));
				
				// ### load the result (if any and if used) from the current stack to the previous expression stack
				if (sym->type.getSize() > 0 && useReturnValue) {
					call->copyResult = gen.getCodeSize();
					gen.genAsm(SymbolicInstr(Compiler::LOAD, Constant(0), INT_CODE, SymbolicInstr::OPERAND, SymbolicInstr::LOCAL));
				}
				
				// ### restore the previous SP
				call->restoreSP = gen.getCodeSize();
				gen.genAsm(SymbolicInstr(Compiler::SETSP, Constant(0)));

				gen.fnCall.push_back(call);

				type = sym->type;
			}
	)
;

//--------------------------------------------------------------------------------------

args
returns [vector<TypeAttr> argTypes]
{
	TypeAttr t;
}:

	#(ARGS
		(
			t = expression

				{
					argTypes.push_back(t);
				}
		)+
	)
;

//--------------------------------------------------------------------------------------

reference[bool isLValue = false, bool genZeroOffset = false]
returns [TypeAttr type]:

	#(n:REF type = expression)
	
		{
			ERROR_IF(! type.isPointer(), n->getLine(), "Cannot dereference a non-pointer value");

			// For an L-value we don't want to push the value onto stack - we just
			// want to set it to something, so we don't generate a LOAD instruction
			// in those cases.
			//
			if (! isLValue) {
				gen.genAsm(SymbolicInstr(Compiler::LOADI0, type.getStorageClass()));
			}

			// this is used for INCI/DECI instructions
			if (genZeroOffset) {
				gen.genAsm(SymbolicInstr(Compiler::ICLOAD, Constant(0)));
			}
			
			type.reduceRefLevel();
		}
;

//--------------------------------------------------------------------------------------

primaryExpr
returns [TypeAttr type]
{
	Symbol* symbol;
	Constant cns;
}:

		symbol = identifier

			{
				type = symbol->type;
			}

	|	type = postfixExpr
	
	|	cns = constant
	
			{
				type = cns.type;
			}
;
	
//--------------------------------------------------------------------------------------

constant
returns [Constant cns]:

	#(CONST
		(
				n:NUMERIC 
					
					{
						cns = getNumericConstant(n->getText());
						scope.top()->addConstant(cns);
						gen.genConst(cns);
					}
			
			|	s:STRING
			
					{
						string str = s->getText();
						
						TypeAttr::DimVector dims;

						// subtract 2 for quotes and add 1 for '\0' terminator
						dims.push_back(static_cast<int>(str.size() - 1));
						
						TypeAttr t(Compiler::CHAR, dims);
						t.setAsPointer();
						
						cns.token = str;
						cns.type = t;
						scope.top()->addConstant(cns);
						
						gen.genConst(cns);
					}
			
			|	c:CHAR 

					{
						cns = Constant(c->getText()[0], TypeAttr(Compiler::CHAR));
						scope.top()->addConstant(cns);
						gen.genConst(cns);
					}
			
			|	TRUE 
			
					{
						cns = Constant(1, TypeAttr(BOOL));
						gen.genConst(cns);
					}

			|	FALSE 
			
					{
						cns = Constant(0, TypeAttr(BOOL));
						gen.genConst(cns);
					}

			|	NULLV

					{
						cns = Constant(0, TypeAttr(INT));
						cns.type.setAsPointer();
						gen.genConst(cns);
					}
		)
	)
;

//--------------------------------------------------------------------------------------

postfixExpr
returns [TypeAttr type]:

	type = arrayAccess
	
		{
			gen.genAsm(SymbolicInstr(Compiler::LOADI, type.getStorageClass(type.isPointer())));
		}
;

//--------------------------------------------------------------------------------------

arrayAccess
returns [TypeAttr arrayType]
{
	vector<int> arrayLayoutValues;
	size_t lv;
	bool generateAdd = false;
	TypeAttr indexType;
	int dimension = 0;
}:

	#(b:LBRACKET 
		arrayType = expression 

			{
				if (!arrayType.isPointer()) {
					gen.setArrayLayoutValues(arrayType, &arrayLayoutValues);
					lv = arrayLayoutValues.size() - 1;
				}
			}

		(
			indexType = expression
			
				{
					ERROR_IF(indexType != TypeAttr(INT), b->getLine(), "Array index must be of integer type");
				
					ERROR_IF(! (arrayType.isArray() || arrayType.isIndexablePointer()), b->getLine(), "Illegal array access");

					ERROR_IF(arrayType.isPointer() && dimension + 1 > 1, b->getLine(),
								"Pointers cannot be treated as multi-dimensional arrays");

					++dimension;

					if (!arrayType.isPointer() && arrayType.isArray()) {
						int layoutValue = arrayLayoutValues[lv];
						if (layoutValue > 1) {
							Constant c(layoutValue, TypeAttr(INT));
							gen.genAsm(SymbolicInstr(Compiler::ICLOAD, c));
							gen.genAsm(SymbolicInstr(Compiler::MUL, INT_CODE));
						}
						
						if (generateAdd) {
							gen.genAsm(SymbolicInstr(Compiler::ADD, INT_CODE));
						}
						
						generateAdd = true;
						--lv;
					}
					
					arrayType.reduceArray();
				}
		)+ 
	)
;

//--------------------------------------------------------------------------------------

identifier[bool genCode = true]
returns [Symbol* sym = 0]:

	#(IDENT id:IDENT)

		{
			sym = findSymbol(id->getText(), scope.getStack());
			
			ERROR_IF(! sym, id->getLine(), "Undefined identifier");
			
			if (genCode) {
				if (sym->type.isArray()) {
					if (sym->isParameter) {
						gen.genAsm(SymbolicInstr(Compiler::LOAD, sym, INT_CODE));	// array parameters are treated as pointers
					}
					else {
						gen.genAsm(SymbolicInstr(Compiler::ICLOAD, sym));		// load the base address of the array
					}
				}
				else {
					gen.genAsm(SymbolicInstr(Compiler::LOAD, sym, sym->type.getStorageClass()));
				}
			}
		}
;
